from dataclasses import dataclass
import datetime
import functools
import inspect
import json
from typing import Callable, Optional
import openai  # noqa
from openai.api_resources import (ChatCompletion, Completion, Edit, Embedding,
                                  Image, Moderation)

from helicone.async_logger.async_logger import (HeliconeAsyncLogger,
                                                HeliconeAyncLogRequest,
                                                Provider, ProviderRequest,
                                                ProviderResponse, Timing)


@dataclass
class HeliconeMeta:
    custom_properties: Optional[dict] = None
    user_id: Optional[str] = None

    def build(self) -> dict:
        meta = {}
        if self.custom_properties:
            for key, value in self.custom_properties.items():
                meta[f"Helicone-Property-{key}"] = value

        if self.user_id:
            meta["Helicone-User-Id"] = self.user_id
        print("METAAAAA", meta)
        return meta


class CreateArgsExtractor:
    _helicone_meta: Optional[HeliconeMeta]

    def __init__(self,
                 api_key=None,
                 api_base=None,
                 api_type=None,
                 request_id=None,
                 api_version=None,
                 organization=None,
                 **kwargs):
        self.kwargs = kwargs
        self.kwargs["api_key"] = api_key
        self.kwargs["api_base"] = api_base
        self.kwargs["api_type"] = api_type
        self.kwargs["request_id"] = request_id
        self.kwargs["api_version"] = api_version
        self.kwargs["organization"] = organization
        self._helicone_meta = kwargs.get("helicone_meta")
        self.kwargs.pop("helicone_meta", None)

    def get_args(self):
        return self.kwargs

    def get_body(self):
        return self.kwargs

    def get_helicone_meta(self) -> dict:
        if (self._helicone_meta is None):
            return {}
        return self._helicone_meta.build()


class OpenAIInjector:
    def __init__(self):
        pass

    def update_response_headers(self, result, helicone_request_id):
        result["helicone_request_id"] = helicone_request_id

    def _result_interceptor(self,
                            result,
                            helicone_meta: dict,
                            send_response: Callable[[dict], None] = None):

        def generator_intercept_packets():
            response = {}
            response["streamed_data"] = []
            for r in result:
                r["helicone_meta"] = helicone_meta
                response["streamed_data"].append(json.loads(r.__str__()))
                yield r
            send_response(response)

        if inspect.isgenerator(result):
            return generator_intercept_packets()
        else:
            result["helicone_meta"] = helicone_meta
            send_response(json.loads(result.__str__()))

            return result

    def _result_interceptor_async(self,
                                  result,
                                  helicone_meta: dict,
                                  send_response: Callable[[dict], None] = None):

        async def generator_intercept_packets():
            response = {}
            response["streamed_data"] = []
            for r in result:
                r["helicone_meta"] = helicone_meta
                response["streamed_data"].append(json.loads(r.__str__()))
                yield r
            send_response(response)

        if inspect.isgenerator(result):
            return generator_intercept_packets()
        else:
            result["helicone_meta"] = helicone_meta
            send_response(json.loads(result.__str__()))

            return result

    def _with_helicone_auth(self, func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            logger = HeliconeAsyncLogger.from_helicone_global()

            arg_extractor = CreateArgsExtractor(*args, **kwargs)
            now = datetime.datetime.now()

            providerRequest = ProviderRequest(
                url="N/A",
                json=arg_extractor.get_body(),
                meta=arg_extractor.get_helicone_meta()
            )
            try:
                result = func(**arg_extractor.get_args())
            except Exception as e:
                later = datetime.datetime.now()
                async_log = HeliconeAyncLogRequest(
                    providerRequest=providerRequest,
                    providerResponse=ProviderResponse(
                        json={
                            "error": str(e)
                        },
                        status=500,
                        headers={
                            "openai-version": "ligmaligma"
                        }
                    ),
                    timing=Timing.from_datetimes(now, later)
                )
                logger.log(async_log, Provider.OPENAI)

                raise e

            def send_response(response):
                later = datetime.datetime.now()
                async_log = HeliconeAyncLogRequest(
                    providerRequest=providerRequest,
                    providerResponse=ProviderResponse(
                        json=response,
                        status=200,
                        headers={
                            "openai-version": "ligmaligma"
                        }

                    ),
                    timing=Timing.from_datetimes(now, later)
                )
                print("logging", async_log, Provider.OPENAI)

                logger.log(async_log, Provider.OPENAI)

            return self._result_interceptor(result,
                                            {},
                                            send_response)

        return wrapper

    def _with_helicone_auth_async(self, func):
        @functools.wraps(func)
        async def wrapper(*args, **kwargs):
            logger = HeliconeAsyncLogger.from_helicone_global()

            arg_extractor = CreateArgsExtractor(*args, **kwargs)
            now = datetime.datetime.now()

            providerRequest = ProviderRequest(
                url="N/A",
                json=arg_extractor.get_body(),
                meta=arg_extractor.get_helicone_meta()
            )
            try:
                result = await func(**arg_extractor.get_args())
            except Exception as e:
                later = datetime.datetime.now()
                async_log = HeliconeAyncLogRequest(
                    providerRequest=providerRequest,
                    providerResponse=ProviderResponse(
                        json={
                            "error": str(e)
                        },
                        status=500,
                        headers={
                            "openai-version": "ligmaligma"
                        }
                    ),
                    timing=Timing.from_datetimes(now, later)
                )
                logger.log(async_log, Provider.OPENAI)

                raise e

            def send_response(response):
                later = datetime.datetime.now()
                async_log = HeliconeAyncLogRequest(
                    providerRequest=providerRequest,
                    providerResponse=ProviderResponse(
                        json=response,
                        status=200,
                        headers={
                            "openai-version": "ligmaligma"
                        }

                    ),
                    timing=Timing.from_datetimes(now, later)
                )
                print("logging", async_log, Provider.OPENAI)

                logger.log(async_log, Provider.OPENAI)

            return self._result_interceptor_async(result,
                                                  {},
                                                  send_response)

        return wrapper

    def apply_helicone_auth(self_parent):
        api_resources_classes = [
            (ChatCompletion, "create", "acreate"),
            (Completion, "create", "acreate"),
            (Edit, "create", "acreate"),
            (Embedding, "create", "acreate"),
            (Image, "create", "acreate"),
            (Moderation, "create", "acreate"),
        ]

        for api_resource_class, method, async_method in api_resources_classes:
            create_method = getattr(api_resource_class, method)
            setattr(api_resource_class, method,
                    self_parent._with_helicone_auth(create_method))

            async_create_method = getattr(api_resource_class, async_method)
            setattr(api_resource_class, async_method,
                    self_parent._with_helicone_auth_async(async_create_method))


injector = OpenAIInjector()
injector.apply_helicone_auth()



import dataclasses
import datetime
import os
import time
from dataclasses import dataclass
from enum import Enum
from typing import Optional

import requests

from helicone.globals.helicone import helicone_global


@dataclass
class ProviderRequest:
    url: str
    json: dict
    meta: dict


@dataclass
class ProviderResponse:
    json: dict
    status: int
    headers: dict


@dataclass
class UnixTimeStamp:
    seconds: int
    milliseconds: int

    @staticmethod
    def from_datetime(dt: datetime) -> 'UnixTimeStamp':
        timestamp = dt.timestamp()
        seconds = int(timestamp)
        milliseconds = int((timestamp - seconds) * 1000)
        return UnixTimeStamp(seconds, milliseconds)


@dataclass
class Timing:
    startTime: UnixTimeStamp
    endTime: UnixTimeStamp

    @staticmethod
    def from_datetimes(start: datetime, end: datetime) -> 'Timing':
        start_timestamp = UnixTimeStamp.from_datetime(start)
        end_timestamp = UnixTimeStamp.from_datetime(end)
        return Timing(start_timestamp, end_timestamp)


@dataclass
class HeliconeAyncLogRequest:
    providerRequest: ProviderRequest
    providerResponse: ProviderResponse
    timing: Timing


@dataclass
class HeliconeMeta:
    custom_properties: Optional[dict]
    user_id: Optional[str]


class Provider(Enum):
    OPENAI = "openai"
    AZURE_OPENAI = "azure-openai"
    ANTHROPIC = "anthropic"


class HeliconeAsyncLogger:
    base_url: str

    def __init__(self,
                 base_url: Optional[str] = None,
                 api_key:  Optional[str] = None,
                 ) -> None:
        self.base_url = base_url or "https://api.hconeai.com"
        self.api_key = api_key or os.environ.get("HELICONE_API_KEY")

    @staticmethod
    def from_helicone_global() -> 'HeliconeAsyncLogger':
        return HeliconeAsyncLogger(
            base_url=helicone_global.base_url,
            api_key=helicone_global.api_key
        )

    def _request(self, body: dict, url: str) -> requests.Response:
        res = requests.post(
            url=url,
            json=body,
            headers={
                "Authorization": f"Bearer {self.api_key}"
            }
        )
        if (res.status_code != 200):
            raise ValueError(
                f"Failed to log to {url}. Status code {res.status_code}")
        return res

    def log(self, request: HeliconeAyncLogRequest,
            provider: Provider,
            meta: Optional[HeliconeMeta] = None
            ):
        print("logging", request, provider)
        if provider == Provider.OPENAI:
            self._request(
                body=dataclasses.asdict(request),
                url=f"{self.base_url}/oai/v1/log"
            )
        elif provider == Provider.AZURE_OPENAI:
            self._request(
                url=f"{self.base_url}/oai/v1/log",
                body=dataclasses.asdict(request),
            )
        elif provider == Provider.ANTHROPIC:
            self._request(
                url=f"{self.base_url}/anthropic/v1/log",
                body=dataclasses.asdict(request),
            )
        else:
            raise ValueError(f"Unknown provider {provider}")



import json
import os
import lockfile


class HeliconeLockManager:
    LOCKFILE_PATH = os.path.expanduser(
        "~/.helicone/custom_properties.json.lock")
    JSON_PATH = os.path.expanduser("~/.helicone/custom_properties.json")

    @staticmethod
    def write_custom_property(property_name, value):
        with lockfile.LockFile(HeliconeLockManager.LOCKFILE_PATH):
            if os.path.exists(HeliconeLockManager.JSON_PATH):
                with open(HeliconeLockManager.JSON_PATH, "r") as json_file:
                    properties = json.load(json_file)
            else:
                properties = {}

            properties[property_name] = value

            with open(HeliconeLockManager.JSON_PATH, "w") as json_file:
                json.dump(properties, json_file)

    @staticmethod
    def clear_all_properties():
        with lockfile.LockFile(HeliconeLockManager.LOCKFILE_PATH):
            with open(HeliconeLockManager.JSON_PATH, "w") as json_file:
                json.dump({}, json_file)

    @staticmethod
    def remove_custom_property(property_name):
        with lockfile.LockFile(HeliconeLockManager.LOCKFILE_PATH):
            if os.path.exists(HeliconeLockManager.JSON_PATH):
                with open(HeliconeLockManager.JSON_PATH, "r") as json_file:
                    properties = json.load(json_file)

                if property_name in properties:
                    del properties[property_name]

                    with open(HeliconeLockManager.JSON_PATH, "w") as json_file:
                        json.dump(properties, json_file)





import logging
from typing import Optional
import os


logger = logging.getLogger(__name__)


class HeliconeGlobal:
    _api_key: Optional[str]
    _base_url: Optional[str]
    _proxy_url: Optional[str]

    def __init__(self,
                 api_key: Optional[str] = None,
                 base_url: Optional[str] = None,
                 proxy_url: Optional[str] = None
                 ):
        self._api_key = api_key
        self._base_url = base_url
        self._proxy_url = proxy_url

    @property
    def api_key(self) -> Optional[str]:
        if (self._api_key is None):
            return os.environ.get("HELICONE_API_KEY")

        return self._api_key

    @api_key.setter
    def api_key(self, value: Optional[str]):
        self._api_key = value

    @property
    def base_url(self) -> Optional[str]:
        if (self._base_url is None):
            return "https://api.hconeai.com"
        return self._base_url

    @base_url.setter
    def base_url(self, value: Optional[str]):
        self._base_url = value

    @property
    def proxy_url(self) -> Optional[str]:
        if (self._proxy_url is None):
            return "http://oai.hconeai.com/v1"
        return self._proxy_url

    @proxy_url.setter
    def proxy_url(self, value: Optional[str]):
        self._proxy_url = value


helicone_global = HeliconeGlobal()



import os
from dotenv import load_dotenv
from helicone.openai_proxy import openai, log_feedback
from helicone.globals import helicone_global
import uuid
from supabase import create_client
import hashlib
import pytest

helicone_global.proxy_url = "http://127.0.0.1:8787/v1"
helicone_global.api_key = os.getenv("HELICONE_API_KEY_LOCAL")

SUPABASE_SERVICE_ROLE_KEY = "eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJpc3MiOiJzdXBhYmFzZS1kZW1vIiwicm9sZSI6InNlcnZpY2Vfcm9sZSIsImV4cCI6MTk4MzgxMjk5Nn0.EGIM96RAZx35lJzdJsyH-qQwv8Hdp7fsn3W0YpN81IU"
SUPABASE_URL = "http://localhost:54321"

sb = create_client(SUPABASE_URL, SUPABASE_SERVICE_ROLE_KEY)

load_dotenv()


def test_cache_completion():
    unique_id = str(uuid.uuid4())
    prompt = f"Cache test with UUID: {unique_id}"

    response1 = openai.Completion.create(
        model="text-ada-001",
        prompt=prompt,
        max_tokens=10,
        cache=True
    )
    assert response1.helicone.cache == "MISS"

    response2 = openai.Completion.create(
        model="text-ada-001",
        prompt=prompt,
        max_tokens=10,
        cache=True
    )
    assert response2.helicone.cache == "HIT"

    response1_copy = response1.copy()
    response2_copy = response2.copy()

    del response1_copy['helicone']['cache']
    del response2_copy['helicone']['cache']

    assert response1_copy == response2_copy


def test_cache_embedding():
    unique_id = str(uuid.uuid4())
    prompt = f"Cache test with UUID: {unique_id}"

    response1 = openai.Embedding.create(
        model="text-embedding-ada-002",
        input=prompt,
        cache=True
    )
    assert response1.helicone.cache == "MISS"

    response2 = openai.Embedding.create(
        model="text-embedding-ada-002",
        input=prompt,
        cache=True
    )
    assert response2.helicone.cache == "HIT"

    response1_copy = response1.copy()
    response2_copy = response2.copy()

    del response1_copy['helicone']['cache']
    del response2_copy['helicone']['cache']

    assert response1_copy == response2_copy


def test_rate_limit_policy():
    rate_limit_policy_dict = {"quota": 10, "time_window": 60}
    rate_limit_policy_str = "10;w=60"

    response = openai.ChatCompletion.create(
        model="gpt-3.5-turbo",
        messages=[{"role": "user", "content": "Rate limit policy test"}],
        rate_limit_policy=rate_limit_policy_dict
    )
    assert response.helicone.rate_limit.policy.startswith(
        rate_limit_policy_str)

    openai.ChatCompletion.create(
        model="gpt-3.5-turbo",
        messages=[{"role": "user", "content": "Rate limit policy test"}],
        rate_limit_policy=rate_limit_policy_str
    )
    # Assert the prefix of the policy is equal to the str
    assert response.helicone.rate_limit.policy.startswith(
        rate_limit_policy_str)


def test_custom_properties():
    properties = {
        "Session": "24",
        "Conversation": "support_issue_2",
        "App": "mobile",
    }

    openai.Completion.create(
        model="text-ada-001",
        prompt="Custom properties test",
        max_tokens=10,
        properties=properties
    )


def hash(key: str) -> str:
    # Encode the key as bytes
    key_bytes = key.encode('utf-8')

    # Create a SHA-256 hash object
    sha256 = hashlib.sha256()

    # Update the hash object with the key bytes
    sha256.update(key_bytes)

    # Get the hexadecimal digest of the hash
    hashed_key_hex = sha256.hexdigest()

    return hashed_key_hex


def fetch_feedback(helicone_id):
    api_key_hash = hash(f'Bearer {os.getenv("HELICONE_API_KEY_LOCAL")}')

    # Fetch the response with the corresponding helicone_id
    response_query = sb.table("response").select(
        "id, request").eq("request", helicone_id)
    response_result = response_query.single().execute()
    response_data = response_result.data

    # Fetch the request with the corresponding response.request value
    request_query = sb.table("request").select(
        "id, helicone_api_keys (id, api_key_hash)").eq("id", response_data["request"])
    request_result = request_query.single().execute()
    request_data = request_result.data

    matching_api_key_hash = request_data["helicone_api_keys"]["api_key_hash"]
    matching_api_key_id = request_data["helicone_api_keys"]["id"]

    # Check if the apiKeyHash matches the helicone_api_key_id's api_key_hash
    if not request_data or matching_api_key_hash != api_key_hash:
        raise ValueError("Not authorized to fetch feedback.")

    # Fetch feedback_metrics for the given api_key_id
    metric_query = sb.table("feedback_metrics").select("id, name, data_type").eq(
        "helicone_api_key_id", str(matching_api_key_id))
    metric_result = metric_query.execute()
    metric_data = metric_result.data

    # Fetch feedback for each feedback_metric and the response
    feedback_data = []
    for metric in metric_data:
        feedback_query = (
            sb.table("feedback")
            .select("created_by, boolean_value, float_value, string_value, categorical_value")
            .eq("response_id", str(response_data["id"]))
            .eq("feedback_metric_id", str(metric["id"]))
        )
        feedback_result = feedback_query.execute()
        feedback = feedback_result.data

        feedback_data.extend(feedback)

    return feedback_data


@pytest.mark.skip(reason="Skipping until feedback is reimplemented")
def test_log_feedback():
    prompt = "Integration test for logging feedback"

    response = openai.Completion.create(
        model="text-ada-001",
        prompt=prompt,
        max_tokens=10,
    )

    log_feedback(response, "score", True, data_type="boolean")

    helicone_id = response['helicone']['id']
    feedback_data = fetch_feedback(helicone_id)

    assert len(feedback_data) == 1
    assert feedback_data[0]["boolean_value"] is True
    assert feedback_data[0]["float_value"] is None
    assert feedback_data[0]["string_value"] is None
    assert feedback_data[0]["categorical_value"] is None


@pytest.mark.skip(reason="Skipping until feedback is reimplemented")
def test_sync_nostream():
    response = openai.ChatCompletion.create(
        model='gpt-3.5-turbo',
        messages=[{
            'role': 'user',
            'content': "Hello World!"
        }],
        properties={"mode": "Create and stream=False"},
        stream=False
    )

    log_feedback(response, "condition", "create_and_stream_false")

    helicone_id = response['helicone']['id']
    feedback_data = fetch_feedback(helicone_id)

    assert len(feedback_data) == 1
    assert feedback_data[0]["boolean_value"] is None
    assert feedback_data[0]["float_value"] is None
    assert feedback_data[0]["string_value"] == "create_and_stream_false"
    assert feedback_data[0]["categorical_value"] is None


@pytest.mark.skip(reason="Skipping until feedback is reimplemented")
def test_sync_stream():
    from collections import deque

    iterator = openai.ChatCompletion.create(
        model='gpt-3.5-turbo',
        messages=[{
            'role': 'user',
            'content': "Hello World!"
        }],
        properties={"mode": "Create and stream=True"},
        stream=True
    )
    last_chunk = deque(iterator, maxlen=1).pop()

    log_feedback(last_chunk, "condition", "create_and_stream_true")

    helicone_id = last_chunk['helicone']['id']
    feedback_data = fetch_feedback(helicone_id)

    assert len(feedback_data) == 1
    assert feedback_data[0]["boolean_value"] is None
    assert feedback_data[0]["float_value"] is None
    assert feedback_data[0]["string_value"] == "create_and_stream_true"
    assert feedback_data[0]["categorical_value"] is None


@pytest.mark.skip(reason="Skipping until feedback is reimplemented")
@pytest.mark.asyncio
async def test_async_nostream():
    response = (await openai.ChatCompletion.acreate(
        model='gpt-3.5-turbo',
        messages=[{
            'role': 'user',
            'content': "Hello World!"
        }],
        properties={"mode": "Acreate and stream=False"},
        stream=False
    ))

    log_feedback(response, "condition", "acreate_and_stream_false")

    helicone_id = response['helicone']['id']
    feedback_data = fetch_feedback(helicone_id)

    assert len(feedback_data) == 1
    assert feedback_data[0]["boolean_value"] is None
    assert feedback_data[0]["float_value"] is None
    assert feedback_data[0]["string_value"] == "acreate_and_stream_false"
    assert feedback_data[0]["categorical_value"] is None


@pytest.mark.skip(reason="Skipping until feedback is reimplemented")
@pytest.mark.asyncio
async def test_async_stream():
    async for chunk in await openai.ChatCompletion.acreate(
        model='gpt-3.5-turbo',
        messages=[{
            'role': 'user',
            'content': "Hello World!"
        }],
        properties={"mode": "Acreate and stream=True"},
        stream=True
    ):
        continue

    log_feedback(chunk, "condition", "acreate_and_stream_true")

    helicone_id = chunk['helicone']['id']
    feedback_data = fetch_feedback(helicone_id)

    assert len(feedback_data) == 1
    assert feedback_data[0]["boolean_value"] is None
    assert feedback_data[0]["float_value"] is None
    assert feedback_data[0]["string_value"] == "acreate_and_stream_true"
    assert feedback_data[0]["categorical_value"] is None


def test_sync_nostream_cache():
    unique_id = str(uuid.uuid4())
    message = f"Sync NoStream Cache test with UUID: {unique_id}"

    response1 = openai.ChatCompletion.create(
        model='gpt-3.5-turbo',
        messages=[{
            'role': 'user',
            'content': message
        }],
        stream=False,
        cache=True
    )
    assert response1.helicone.cache == "MISS"

    response2 = openai.ChatCompletion.create(
        model='gpt-3.5-turbo',
        messages=[{
            'role': 'user',
            'content': message
        }],
        stream=False,
        cache=True
    )
    assert response2.helicone.cache == "HIT"

    response1_copy = response1.copy()
    response2_copy = response2.copy()

    del response1_copy['helicone']['cache']
    del response2_copy['helicone']['cache']

    assert response1_copy == response2_copy


@pytest.mark.asyncio
async def test_async_nostream_cache():
    unique_id = str(uuid.uuid4())
    message = f"Async NoStream Cache test with UUID: {unique_id}"

    response1 = (await openai.ChatCompletion.acreate(
        model='gpt-3.5-turbo',
        messages=[{
            'role': 'user',
            'content': message
        }],
        stream=False,
        cache=True
    ))
    assert response1.helicone.cache == "MISS"

    response2 = (await openai.ChatCompletion.acreate(
        model='gpt-3.5-turbo',
        messages=[{
            'role': 'user',
            'content': message
        }],
        stream=False,
        cache=True
    ))
    assert response2.helicone.cache == "HIT"

    response1_copy = response1.copy()
    response2_copy = response2.copy()

    del response1_copy['helicone']['cache']
    del response2_copy['helicone']['cache']

    assert response1_copy == response2_copy


def test_azure():
    openai.api_type = "azure"
    openai.api_base = f"https://{os.environ['AZURE_OPENAI_ENDPOINT']}"
    openai.api_version = "2023-03-15-preview"
    openai.api_key = os.environ["AZURE_OPENAI_API_KEY"]

    openai.ChatCompletion.create(
        engine="gpt-4",
        messages=[
            {
                "role": "user",
                "content": "Say hi!",
            },
        ],
        temperature=1,
        max_tokens=10,
    )



import functools
import uuid
import inspect
from helicone.globals import helicone_global
import openai
import requests
from openai.api_resources import (
    ChatCompletion,
    Completion,
    Edit,
    Embedding,
    Image,
    Moderation,
)
import logging
import threading

logger = logging.getLogger(__name__)

global_headers = {}


class AttributeDict(dict):
    def __init__(self, *args, **kwargs):
        super(AttributeDict, self).__init__(*args, **kwargs)
        self.__dict__ = self


def normalize_data_type(data_type):
    if isinstance(data_type, str):
        data_type = data_type.lower()

    if data_type in (str, "str", "string"):
        return "string"
    elif data_type in (bool, "bool", "boolean"):
        return "boolean"
    elif data_type in (float, int, "float", "int", "numerical"):
        return "numerical"
    elif data_type in (object, "object", "categorical"):
        return "categorical"
    else:
        raise ValueError(
            "Invalid data_type provided. Please use a valid data type or string.")


def prepare_api_base(**kwargs):
    original_api_base = openai.api_base
    kwargs["headers"].update({"Helicone-OpenAI-Api-Base": original_api_base})

    openai.api_base = helicone_global.proxy_url

    if openai.api_type == "azure":
        if helicone_global.proxy_url.endswith('/v1'):
            if helicone_global.proxy_url != "https://oai.hconeai.com/v1":
                logging.warning(
                    f"Detected likely invalid Azure API URL when proxying Helicone with proxy url {helicone_global.proxy_url}. Removing '/v1' from the end.")
            openai.api_base = helicone_global.proxy_url[:-3]

    return original_api_base, kwargs


class Helicone:
    def __init__(self):
        self.openai = openai
        self.apply_helicone_auth()
        self.headers_store = {}

    def log_feedback(self, response, name, value, data_type=None):
        helicone_id = response.get("helicone", {}).get("id")
        if not helicone_id:
            raise ValueError(
                "The provided response does not have a valid Helicone ID.")

        feedback_data = {
            "helicone-id": helicone_id,
            "name": name,
            "value": value,
        }
        if data_type:
            feedback_data["data-type"] = normalize_data_type(data_type)

        url = f"{helicone_global.proxy_url}/feedback"

        headers = {
            "Content-Type": "application/json",
            "Helicone-Auth": f"Bearer {helicone_global.api_key}",
        }

        response = requests.post(url, headers=headers, json=feedback_data)
        if response.status_code != 200:
            logger.error(f"HTTP error occurred: {response.status_code}")
            logger.error(
                f"Response content: {response.content.decode('utf-8', 'ignore')}")

            response.raise_for_status()
        return response.json()

    def _prepare_headers(self, **kwargs):
        headers = kwargs.get("headers", {})

        if "Helicone-Auth" not in headers and helicone_global.api_key:
            headers["Helicone-Auth"] = f"Bearer {helicone_global.api_key}"

        # Generate a UUID and add it to the headers
        helicone_request_id = str(uuid.uuid4())
        headers["helicone-request-id"] = helicone_request_id

        headers.update(self._get_property_headers(
            kwargs.pop("properties", {})))
        headers.update(self._get_cache_headers(kwargs.pop("cache", None)))
        headers.update(self._get_retry_headers(kwargs.pop("retry", None)))
        headers.update(self._get_rate_limit_policy_headers(
            kwargs.pop("rate_limit_policy", None)))

        kwargs["headers"] = headers

        return helicone_request_id, kwargs

    def update_response_headers(self, result, helicone_request_id):
        headers = self.headers_store.get(helicone_request_id, {})
        result["helicone"] = AttributeDict(
            id=headers.get("Helicone-Id"),
            status=headers.get("Helicone-Status"),
            cache=headers.get("Helicone-Cache"),
            rate_limit=AttributeDict(
                limit=headers.get("Helicone-RateLimit-Limit"),
                remaining=headers.get("Helicone-RateLimit-Remaining"),
                reset=headers.get("Helicone-RateLimit-Reset"),
                policy=headers.get("Helicone-RateLimit-Policy"),
            ) if headers.get("Helicone-RateLimit-Policy") else None,
        )

    def _modify_result(self, result, helicone_request_id):
        def result_with_helicone():
            for r in result:
                self.update_response_headers(r, helicone_request_id)
                yield r

        if inspect.isgenerator(result):
            return result_with_helicone()
        else:
            self.update_response_headers(result, helicone_request_id)
            return result

    async def _modify_result_async(self, result, helicone_request_id):
        async def result_with_helicone_async():
            async for r in result:
                self.update_response_headers(r, helicone_request_id)
                yield r

        if inspect.isasyncgen(result):
            return result_with_helicone_async()
        else:
            self.update_response_headers(result, helicone_request_id)
            return result

    def _with_helicone_auth(self, func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            helicone_request_id, kwargs = self._prepare_headers(**kwargs)
            original_api_base, kwargs = prepare_api_base(**kwargs)

            try:
                result = func(*args, **kwargs)
            finally:
                openai.api_base = original_api_base

            return self._modify_result(result, helicone_request_id)

        return wrapper

    def _with_helicone_auth_async(self, func):
        @functools.wraps(func)
        async def wrapper(*args, **kwargs):
            helicone_request_id, kwargs = self._prepare_headers(**kwargs)
            original_api_base, kwargs = prepare_api_base(**kwargs)

            try:
                result = await func(*args, **kwargs)
            finally:
                openai.api_base = original_api_base

            return await self._modify_result_async(result, helicone_request_id)

        return wrapper

    def _get_property_headers(self, properties):
        return {f"Helicone-Property-{key}": str(value) for key, value in properties.items()}

    def _get_cache_headers(self, cache):
        return {"Helicone-Cache-Enabled": "true"} if cache is True else {}

    def _get_retry_headers(self, retry):
        if isinstance(retry, bool) and retry:
            return {"Helicone-Retry-Enabled": "true"}
        elif isinstance(retry, dict):
            headers = {"Helicone-Retry-Enabled": "true"}
            if "num" in retry:
                headers["Helicone-Retry-Num"] = str(retry["num"])
            if "factor" in retry:
                headers["Helicone-Retry-Factor"] = str(retry["factor"])
            if "min_timeout" in retry:
                headers["Helicone-Retry-Min-Timeout"] = str(
                    retry["min_timeout"])
            if "max_timeout" in retry:
                headers["Helicone-Retry-Max-Timeout"] = str(
                    retry["max_timeout"])
            return headers
        return {}

    def _get_rate_limit_policy_headers(self, rate_limit_policy):
        if rate_limit_policy:
            if isinstance(rate_limit_policy, str):
                policy = rate_limit_policy
            elif isinstance(rate_limit_policy, dict):
                policy = f'{rate_limit_policy["quota"]};w={rate_limit_policy["time_window"]}'
                if "segment" in rate_limit_policy:
                    policy += f';s={rate_limit_policy["segment"]}'
            else:
                raise TypeError(
                    "rate_limit_policy must be either a string or a dictionary")
            return {"Helicone-RateLimit-Policy": policy}
        return {}

    def apply_helicone_auth(self_parent):
        def request_raw_patched(self, *args, **kwargs):
            helicone_id = kwargs["supplied_headers"]["helicone-request-id"]
            response = original_request_raw(self, *args, **kwargs)
            if helicone_id:
                with threading.Lock():
                    self_parent.headers_store[helicone_id] = response.headers
            return response

        async def arequest_raw_patched(self, *args, **kwargs):
            helicone_id = kwargs["supplied_headers"]["helicone-request-id"]
            response = await original_arequest_raw(self, *args, **kwargs)
            if helicone_id:
                with threading.Lock():
                    self_parent.headers_store[helicone_id] = response.headers
            return response

        original_request_raw = openai.api_requestor.APIRequestor.request_raw
        openai.api_requestor.APIRequestor.request_raw = request_raw_patched

        original_arequest_raw = openai.api_requestor.APIRequestor.arequest_raw
        openai.api_requestor.APIRequestor.arequest_raw = arequest_raw_patched

        api_resources_classes = [
            (ChatCompletion, "create", "acreate"),
            (Completion, "create", "acreate"),
            (Edit, "create", "acreate"),
            (Embedding, "create", "acreate"),
            (Image, "create", "acreate"),
            (Moderation, "create", "acreate"),
        ]

        for api_resource_class, method, async_method in api_resources_classes:
            create_method = getattr(api_resource_class, method)
            setattr(api_resource_class, method,
                    self_parent._with_helicone_auth(create_method))

            async_create_method = getattr(api_resource_class, async_method)
            setattr(api_resource_class, async_method,
                    self_parent._with_helicone_auth_async(async_create_method))


helicone = Helicone()
log_feedback = helicone.log_feedback
